
{# TODO: Fix short description without JS #}
{# Display description #}
{#{%- if description -%}#}
{#    {%- if description is description_short -%}#}
{#        <span class="description">{{ description }}</span>#}
{#    {%- else -%}#}
{#        <div class="description collapse" id="collapseDescription_{{ schema.html_id }}">#}
{{ description }}
{#        </div>#}
{#        <div><a class="collapse-description-link collapsed" href="#collapseDescription_{{ schema.html_id }}"></a></div>#}
{#    {%- endif -%}#}
{#{%- endif -%}#}
